/* Goewie.java - Probe burp GUI
 *
 * Copyright (c) 2017 Eddy Vanlerberghe.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of Eddy Vanlerberghe shall not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EDDY VANLERBERGHE ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL EDDY VANLERBERGHE BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package evilwan;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.util.*;

/**
 * Probe burp GUI
 */
public class Goewie extends JPanel implements ActionListener {
    // {{ Class members }}

    private static JTable _websockets_table = null;
    
    // {{ Instance members }}

    private ArrayList _table_data = null;
    private boolean   _stuff_initialized = false;
    private boolean   _stuff_dumped = false;
    private JButton	_btn = null;    

    // {{ Class methods }}

    /**
     * Helper method for printing messages.
     * <p>
     * The printed message is written to <code>System.out</code> and starts with the current
     * class name, followed by two dashes, followed by the specified text.
     * @param s contains the <code>String</code> value to print.
     */
    private static void say(String s) {
	System.out.println("Goewie -- " + s);
    }
    
    private static String b2a(byte b) {
	return String.format("0x%02x", ((int)b) & 0xff);
    }
    private static String bar2a(byte[] bar) {
	StringBuilder sb = new StringBuilder();
	StringBuilder asc = new StringBuilder();
	sb.append('[');
	int l = bar.length;
	int n = l;
	if(l > 16) {
	    l = 16;
	}
	for(int j = 0; j<l; ++j) {
	    if(j > 0) {
		sb.append(',');
		sb.append(' ');
	    }
	    sb.append(b2a(bar[j]));
	    int c = ((int)bar[j]) & 0xff;
	    if((c >= 0x20) && (c <= 0xfe)) {
		asc.append((char) c);
	    } else {
		asc.append('.');
	    }
	}
	sb.append(']');
	sb.append('<');
	sb.append("" + n);
	sb.append(':');
	sb.append(asc.toString());
	sb.append('>');
	return sb.toString();
    }

    private static String dumpComponent(Component cmp, String ind) {
	StringBuilder sb = new StringBuilder();
	Class c = cmp.getClass();
	sb.append(dumpClassHierarchy(c, ind));
	String nam = cmp.getName();
	if(nam != null) {
	    sb.append("\n" + ind + "    Name: \"" + nam + "\"");
	}
	return sb.toString();
    }
    
    private static String dumpJLabel(JLabel lab, String ind) {
	StringBuilder sb = new StringBuilder();
	sb.append(dumpComponent(lab, ind));
	sb.append("\n" + ind + "    Text: " + lab.getText());
	return sb.toString();
    }
    
    private static String dumpJButton(JButton btn, String ind) {
	StringBuilder sb = new StringBuilder();
	sb.append(dumpComponent(btn, ind));
	String t = btn.getText();
	if((t != null) && (t.length() > 0)) {
	    sb.append("\n" + ind + "    Text: " + t);
	}
	return sb.toString();
    }
    
    private static String dumpJToggleButton(JToggleButton btn, String ind) {
	StringBuilder sb = new StringBuilder();
	sb.append(dumpComponent(btn, ind));
	String t = btn.getText();
	if((t != null) && (t.length() > 0)) {
	    sb.append("\n" + ind + "    Text: " + t);
	}
	return sb.toString();
    }
    
    private static String dumpJTextField(JTextField txt, String ind) {
	StringBuilder sb = new StringBuilder();
	sb.append(dumpComponent(txt, ind));
	String t = txt.getText();
	if((t != null) && (t.length() > 0)) {
	    sb.append("\n" + ind + "    Text: " + t);
	}
	return sb.toString();
    }
    
    private static String dumpJTextArea(JTextArea txt, String ind) {
	StringBuilder sb = new StringBuilder();
	sb.append(dumpComponent(txt, ind));
	String t = txt.getText();
	if((t != null) && (t.length() > 0)) {
	    sb.append("\n" + ind + "    Text: " + t);
	}
	return sb.toString();
    }
    
    private static String dumpJTabbedPane(JTabbedPane tabs, String ind) {
	StringBuilder sb = new StringBuilder();
	sb.append(dumpComponent(tabs, ind));
	int n = tabs.getTabCount();
	sb.append("\n" + ind + "    Tabs: ");
	for(int i = 0; i < n; ++i) {
	    if(i > 0) {
		sb.append(", ");
	    }
	    sb.append('\"');
	    sb.append(tabs.getTitleAt(i));
	    sb.append('\"');
	}
	sb.append(dumpKids(tabs, ind));
	return sb.toString();
    }
    
    private static String dumpJPanel(JPanel panel, String ind) {
	StringBuilder sb = new StringBuilder();
	sb.append(dumpComponent(panel, ind));
	LayoutManager layout = panel.getLayout();
	if(layout != null) {
	    sb.append("\n" + ind + "    Layout: " + layout.getClass().getName());
	}
	sb.append("\n" + ind + "    Size: " + panel.getWidth() + " x " + panel.getHeight());
	sb.append(dumpKids(panel, ind));
	return sb.toString();
    }
    
    private static String dumpJTable(JTable tab, String ind) {
	StringBuilder sb = new StringBuilder();
	sb.append(dumpComponent(tab, ind));
	TableModel model = tab.getModel();
	sb.append("\n" + ind + "    TableModel: ");
	sb.append(dumpClassHierarchy(model.getClass(), " "));
	int n = model.getColumnCount();
	sb.append("\n" + ind + "    Column headings: ");
	for(int i = 0; i < n; ++i) {
	    if(i > 0) {
		sb.append(", ");
	    }
	    sb.append('\"');
	    String colnam = model.getColumnName(i);
	    sb.append(colnam);
	    sb.append('\"');
	    //
	    // Hack: in current release the only table with a column "Direction" is
	    // the JTable for websocket traffic
	    //
	    if(colnam.equals("Direction")) {
		_websockets_table = tab;
	    }
	}
	//
	// Dump listeners (if any)
	//
	ListSelectionModel cellSelectionModel = tab.getSelectionModel();
	sb.append("\n" + ind + "    ListSelectionModel: ");
	sb.append(dumpClassHierarchy(cellSelectionModel.getClass(), " "));
	ListSelectionListener[] listeners = ((DefaultListSelectionModel) cellSelectionModel).
	    getListSelectionListeners();
	if(listeners != null) {
	    if(listeners.length <= 0) {
		sb.append("\n" + ind + " ListSelectionListener: 0");
	    } else {
		for(int i = 0; i < listeners.length; ++i) {
		    sb.append("\n" + ind + " ListSelectionListener: ");
		    sb.append(dumpClassHierarchy(listeners[i].getClass(),
						 ind + "    "));
		}
	    }
	} else {
	    sb.append("\n" + ind + " ListSelectionListener: <none>");
	}
	//
	// Now proceed as with other components
	//
	sb.append(dumpKids(tab, ind));
	return sb.toString();
    }

    private static String dumpKids(Container box, String ind) {
	StringBuilder sb = new StringBuilder();
	Component[] parts = box.getComponents();
	for(int i = 0; i < parts.length; ++i) {
	    sb.append('\n');
	    if(parts[i] instanceof JLabel) {
		sb.append(dumpJLabel((JLabel) parts[i], ind + "    "));
	    } else if(parts[i] instanceof JButton) {
		sb.append(dumpJButton((JButton) parts[i], ind + "    "));
	    } else if(parts[i] instanceof JToggleButton) {
		sb.append(dumpJToggleButton((JToggleButton) parts[i], ind + "    "));
	    } else if(parts[i] instanceof JTextField) {
		sb.append(dumpJTextField((JTextField) parts[i], ind + "    "));
	    } else if(parts[i] instanceof JTextArea) {
		sb.append(dumpJTextArea((JTextArea) parts[i], ind + "    "));
	    } else if(parts[i] instanceof JTabbedPane) {
		sb.append(dumpJTabbedPane((JTabbedPane) parts[i], ind + "    "));
	    } else if(parts[i] instanceof JTable) {
		sb.append(dumpJTable((JTable) parts[i], ind + "    "));
	    } else if(parts[i] instanceof CellRendererPane) {
		sb.append(dumpCellRendererPane((CellRendererPane) parts[i],
					       ind + "    "));
	    } else if(parts[i] instanceof JPanel) {
		sb.append(dumpJPanel((JPanel) parts[i], ind + "    "));
	    } else if(parts[i] instanceof Container) {
		sb.append(dumpContainer((Container) parts[i], ind + "    "));
	    } else {
		sb.append(dumpComponent(parts[i], ind + "    "));
	    }
	}
	return sb.toString();
    }

    private static String dumpCellRendererPane(CellRendererPane box, String ind) {
	StringBuilder sb = new StringBuilder();
	Class c = box.getClass();
	sb.append(dumpClassHierarchy(c, ind));
	EventListener[] listeners = (EventListener[])
	    (box.getListeners(ListSelectionListener.class));
	if(listeners != null) {
	    if(listeners.length <= 0) {
		sb.append("\n" + ind + "    EventListener: 0");
	    } else {
		for(int i = 0; i < listeners.length; ++i) {
		    sb.append("\n" + ind + "    EventListener: ");
		    sb.append(dumpClassHierarchy(listeners[i].getClass(),
						 ind + "    "));
		}
	    }
	} else {
	    sb.append("\n" + ind + "    EventListener: <none>");
	}
	sb.append(dumpKids(box, ind));
	return sb.toString();
    }

    private static String dumpContainer(Container box, String ind) {
	StringBuilder sb = new StringBuilder();
	Class c = box.getClass();
	sb.append(dumpClassHierarchy(c, ind));
	if(box instanceof Frame) {
	    sb.append("\n" + ind + "    Title: \"" + ((Frame)box).getTitle() + "\"");
	}
	sb.append(dumpKids(box, ind));
	return sb.toString();
    }

    private static String dumpClassHierarchy(Class c, String ind) {
	StringBuilder sb = new StringBuilder();
	sb.append(ind);
	sb.append("Class: ");
	sb.append(c.getName());

	Class mom = c.getSuperclass();
	sb.append(" extends ");
	sb.append(mom.getName());
	while(!mom.getName().startsWith("java")) {
	    mom = mom.getSuperclass();
	    sb.append(" -> ");
	    sb.append(mom.getName());
	}
	
	//Class[] interfaces = c.getInterfaces();
	//if(interfaces.length > 0) {
	//    sb.append(" implements ");
	//    for(int i = 0; i < interfaces.length; ++i) {
	//	if(i > 0) {
	//	    sb.append(", ");
	//	}
	//	sb.append(interfaces[i].getName());
	//    }
	//}
	//if(!mom.getName().equals("java.lang.Object")) {
	if(!mom.getName().startsWith("java")) {
	    sb.append('\n');
	    sb.append(dumpClassHierarchy(mom, ind + "    "));
	}
	return sb.toString();
    }

    private static void dumpFrame(Frame frame) throws Exception {
	String o = "Title: \"" + frame.getTitle() + "\"\n";
	System.out.println("Title: \"" + frame.getTitle() + "\"");
	System.out.println(dumpClassHierarchy(frame.getClass(), "    "));
    }

    public static void dumpFrames() throws Exception {
	Frame[] frames = Frame.getFrames();
	for(int i = 0; i < frames.length; ++i) {
	    //say("Frame [" + i + "] is a " + frames[i].getClass().getName());
	    dumpFrame(frames[i]);
	}
    }

    public static void dumpBurpComponents() throws Exception {
	Frame[] frames = Frame.getFrames();
	for(int i = 0; i < frames.length; ++i) {
	    //if(frames[i].getTitle().startsWith("Burp Suite")) {
	    ////dumpFrame(frames[i]);
	    say("================== Burp GUI tree[" + i + "]:\n" + dumpContainer(frames[i], ""));
	    //}
	}
    }
    public static void slurpBurpComponents() throws Exception {
	Frame[] frames = Frame.getFrames();
	for(int i = 0; i < frames.length; ++i) {
	    //if(frames[i].getTitle().startsWith("Burp Suite")) {
	    ////dumpFrame(frames[i]);
	    say("================== Burp GUI tree[" + i + "]:\n" + dumpContainer(frames[i], ""));
	    //}
	}
    }
    /**
     * Find the main application frame (use its title for locating which Frame)
     */
    public static Frame findTopFrame() throws Exception {
	Frame[] frames = Frame.getFrames();
	for(int i = 0; i < frames.length; ++i) {
	    if(frames[i].getTitle().startsWith("Burp Suite")) {
		return frames[i];
	    }
	}
	return null;
    }
    /**
     * Find the <code>JTabbedPane</code> containing a tab with the specified name.
     * <p>
     * Main reason for using this method is to be able to select the named tab later on.
     */
    public static JTabbedPane findTabbedPane(Container c, String tabnam) throws Exception {
	if(c instanceof JTabbedPane) {
	    JTabbedPane tp = (JTabbedPane) c;
	    int n = tp.getTabCount();
	    for(int i = 0; i < n; ++i) {
		if(tabnam.equals(tp.getTitleAt(i))) {
		    tp.setSelectedIndex(i);
		    return tp;
		}
	    }
	}
	Component[] parts = c.getComponents();
	for(int i = 0; i < parts.length; ++i) {
	    if(parts[i] instanceof Container) {
		JTabbedPane o = findTabbedPane((Container)parts[i], tabnam);
		if(o != null) {
		    return o;
		}
	    }
	}
	return null;
    }
    /**
     * Find component in <code>JTabbedPane</code> that belongs to the specified tab.
     */
    public static Component findNamedTabPane(Container c, String tabnam) throws Exception {
	if(c instanceof JTabbedPane) {
	    JTabbedPane tp = (JTabbedPane) c;
	    int n = tp.getTabCount();
	    for(int i = 0; i < n; ++i) {
		if(tabnam.equals(tp.getTitleAt(i))) {
		    return tp.getComponentAt(i);
		}
	    }
	}
	Component[] parts = c.getComponents();
	for(int i = 0; i < parts.length; ++i) {
	    if(parts[i] instanceof Container) {
		Component o = findNamedTabPane((Container)parts[i], tabnam);
		if(o != null) {
		    return o;
		}
	    }
	}
	return null;
    }
    /**
     * Select named tab in <code>JTabbedPane</code>
     */
    public static int selectNamedTabPane(Container c, String tabnam) throws Exception {
	if(c instanceof JTabbedPane) {
	    JTabbedPane tp = (JTabbedPane) c;
	    int n = tp.getTabCount();
	    for(int i = 0; i < n; ++i) {
		if(tabnam.equals(tp.getTitleAt(i))) {
		    tp.setSelectedIndex(i);
		    return i;
		}
	    }
	}
	Component[] parts = c.getComponents();
	for(int i = 0; i < parts.length; ++i) {
	    if(parts[i] instanceof Container) {
		int o = selectNamedTabPane((Container)parts[i], tabnam);
		if(o < 0) {
		    return -1;
		}
	    }
	}
	return -1;
    }
    /**
     * Find the <code>JLabel</code> component containing the descriptive text for the intercepted message.
     * <p>
     * This code is based on analyzing the GUI components at run-time.
     */
    public static JLabel findMessageLabel(JPanel c) throws Exception {
	JPanel pnlwrk = null;
	Component[] parts = c.getComponents();
	for(int i = 0; i < parts.length; ++i) {
	    if((parts[i] instanceof JPanel) && ((int)parts[i].getBounds().getX() == 0) &&
	       ((int)parts[i].getBounds().getY() == 0)){
		pnlwrk = (JPanel) parts[i];
		break;
	    }
	}
	if(pnlwrk == null) {
	    throw new Exception("Failed to find JPanel at (0,0) in tabbed pane tab");
	}
	JPanel pnl = null;
	parts = pnlwrk.getComponents();
	for(int i = 0; i < parts.length; ++i) {
	    if((parts[i] instanceof JPanel) && ((int)parts[i].getBounds().getX() == 0) &&
	       ((int)parts[i].getBounds().getY() == 0)){
		pnl = (JPanel) parts[i];
		break;
	    }
	}
	if(pnl == null) {
	    throw new Exception("Failed to find JPanel at (0,0) surrounding message JLabel");
	}
	parts = pnl.getComponents();
	for(int i = 0; i < parts.length; ++i) {
	    if(parts[i] instanceof JLabel) {
		 return (JLabel) parts[i];
	    }
	}
	return null;
    }

    /**
     * Look for <code>JButton</code> with a label that contains the specified name.
     */
    public static AbstractButton findNamedButton(Component c, String btnnam) throws Exception {
	if(c instanceof AbstractButton) {
	    AbstractButton btn = (AbstractButton) c;
	    if(btn.getText().indexOf(btnnam) >= 0) {
		return btn;
	    }
	}
	if(c instanceof Container) {
	    Component[] parts = ((Container)c).getComponents();
	    for(int i = 0; i < parts.length; ++i) {
		AbstractButton o = (AbstractButton) findNamedButton(parts[i], btnnam);
		if(o != null) {
		    return o;
		}
	    }
	}
	return null;
    }
    /**
     * Find first <code>JTable</code> in container
     */
    public static JTable findJTable(Container c) throws Exception {
	if(c instanceof JTable) {
	    return (JTable)c;
	}
	Component[] parts = c.getComponents();
	for(int i = 0; i < parts.length; ++i) {
	    if(parts[i] instanceof Container) {
		JTable o = findJTable((Container)parts[i]);
		if(o != null) {
		    return o;
		}
	    }
	}
	return null;
    }

    // {{ Constructors }}
    
    /**
     * Default public constructor.
     */
    public Goewie() {
	setLayout(new BorderLayout());
	_btn = new JButton("Go!");
	_btn.addActionListener(this);
	add(_btn, BorderLayout.CENTER);
    }
    
    // {{ Instance methods }}

    public void actionPerformed(ActionEvent e) {
	//
	// First dump GUI components (and look for JTable containing the websocket
	// data)
	//
	try {
	    dumpBurpComponents();
	    say("Going to analyze: " + _websockets_table);
	    if(_websockets_table != null) {
		TableModel model = _websockets_table.getModel();
		say("~~~~~~~~~~ _websockets_table TableModel properties:\n" +
		    ObjectDumper.dumpTableModelObject(model));
		//Foefel.dumpObjectProps(model, "", new Stack<Object>()));
	    }
	} catch(Exception ex) {
	    say("Caught:" + ex);
	    ex.printStackTrace();
	}
    }
    
}
