/* ObjectDumper.java - Dump internal structure of any given Java object.
 *
 * Copyright (c) 2017 Eddy Vanlerberghe.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of Eddy Vanlerberghe shall not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EDDY VANLERBERGHE ''AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL EDDY VANLERBERGHE BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 *****************************************************************************/
package evilwan;

import java.util.*;
import java.lang.reflect.*;

/**
 * Dump internal structure of any given Java object.
 */
public class ObjectDumper {
    // {{ Class members }}
    private static Hashtable<String, List<Field>> _cache = new Hashtable<String, List<Field>>();

    private static Hashtable<String, String> _ocache = new Hashtable<String, String>();

    // {{ Instance members }}
    
    // {{ Class methods }}
    /**
     * Helper method for printing messages.
     * <p>
     * The printed message is written to <code>System.out</code> and starts with the current
     * class name, followed by two dashes, followed by the specified text.
     * @param s contains the <code>String</code> value to print.
     */
    private static void say(String s) {
	System.out.println("ObjectDumper -- " + s);
    }
    
    /**
     * Retrieve all fields, including those defined in super classes (which are
     * not returned by getDeclaredFields()) but stop before Object class (way too much
     * info that is usually not that interesting for reversing purposes)
     */
    public static List<Field> getAllFields(Class<?> clz, String ind) {
	String clznam = clz.getName();
	//say("getAllFields() " + ind + "-- clz is a " + clz.getName());
	List<Field> flds = new ArrayList<Field>();
	if(clznam.startsWith("java")) {
	    return flds;
	}
	Collections.addAll(flds, clz.getDeclaredFields());
	_cache.put(clznam, flds);
	Class<?> dad = clz.getSuperclass();
	if((dad != null) && !dad.equals(Object.class) &&
	   (_cache.get(dad.getName()) == null)) {
	    flds.addAll(getAllFields(dad, ind + "    "));
	}
	return flds;
    }

    private static Object appendField(StringBuilder sb, Object o, String fldnam) throws Exception {
	Class<?> clz = o.getClass();	// assume never null
	Object oo;
	Field fld = clz.getDeclaredField(fldnam);
	fld.setAccessible(true);	// Make sure we can access the field
	oo = fld.get(o);
	sb.append("\t");
	sb.append(fldnam);
	sb.append("\t= ");
	if(oo == null) {
	    sb.append("<null>");
	} else {
	    sb.append(oo.getClass().getName());
	}
	if(oo instanceof Collection) {
	    Collection ccc = (Collection) oo;
	    sb.append(" <--- Collection: size=");
	    sb.append(ccc.size());
	    if(ccc.size() > 0) {
		Object[] elms = ccc.toArray();
		sb.append(", first element is a ");
		sb.append(elms[0].getClass().getName());
		if(elms[0] instanceof burp.v1h) {
		    burp.v1h que = (burp.v1h) elms[0];
		    Method[] metz = que.getClass().getDeclaredMethods();
		    for(Method mmm : metz) {
			if(mmm.getParameterCount() == 0) {
			    sb.append("\n\t\t");
			    sb.append(mmm.getName());
			    sb.append("() returns a ");
			    String resnam = mmm.getReturnType().getName();
			    sb.append(resnam);
			    sb.append(" = ");
			    if(resnam.equals("int")) {
				sb.append((int) mmm.invoke(elms[0]));
			    } else if(resnam.equals("long")) {
				sb.append((long) mmm.invoke(elms[0]));
			    } else if(resnam.equals("byte")) {
				sb.append((byte) mmm.invoke(elms[0]));
			    } else if(resnam.equals("boolean")) {
				sb.append((boolean) mmm.invoke(elms[0]));
			    } else if(resnam.equals("java.lang.String")) {
				sb.append((String) mmm.invoke(elms[0]));
			    } else {
				Object ooooo = mmm.invoke(elms[0]);
				if(ooooo == null) {
				    sb.append("<null>");
				} else {
				    sb.append(ooooo.toString());
				}
			    }
			}
		    }
		}
	    }
	    sb.append(" --->");
	}
	sb.append("\n");
	return oo;
    }
    
    public static String dumpTableModelObject(Object o) throws Exception {
	Class<?> clz = o.getClass();	// assume never null
	String clznam = clz.getName();
	StringBuilder sb = new StringBuilder();
	sb.append("TableModel: class=");
	sb.append(clznam);
	sb.append("\n");

	Object oo;

	appendField(sb, o, "D");
	oo = appendField(sb, o, "A");
	appendField(sb, o, "w");
	appendField(sb, o, "y");
	appendField(sb, o, "z");
	appendField(sb, o, "x");
	appendField(sb, o, "s");
	appendField(sb, o, "C");
	appendField(sb, o, "t");
	appendField(sb, o, "v");
	appendField(sb, o, "u");
	appendField(sb, o, "B");
	if(false) {
	    Class<?> clzw = oo.getClass();	// assume never null
	    Method getone = clzw.getDeclaredMethod("b", Integer.TYPE);
	    if(getone == null) {
		sb.append("dumpTableModelObject() -- failed to get method for retrieving websocket data...\n");
	    } else {
		sb.append("dumpTableModelObject() -- found method for retrieving websocket data...\n");
		Object ooo = getone.invoke(oo, 0);
		if(ooo == null) {
		    sb.append("dumpTableModelObject() -- failed to get first websocket data...\n");
		} else {
		    sb.append("dumpTableModelObject() -- websocket data is a " +
			      ooo.getClass().getName() + "\n");
		    //
		    // Use accessor methods on that first message
		    //
		    Method[] mets = ooo.getClass().getDeclaredMethods();
		    for(Method mm : mets) {
			sb.append("dumpTableModelObject() -- method: ");
			sb.append(mm.toString());
			sb.append("\n");
		    }
		    /*
		      Method m_c = ooo.getClass().getDeclaredMethod("b");
		      if(m_c == null) {
		      sb.append("dumpTableModelObject() --  could not find message.b() method");
		      } else {
		      Object oooo = m_c.invoke(ooo);
		      if(oooo == null) {
		      sb.append("dumpTableModelObject() --   message.b() returned null");
		      } else {
		      sb.append("dumpTableModelObject() --   message.b() returned a " +
		      oooo.getClass().getName());
		      }
		      }
		    */
		}
	    }
	}
	return sb.toString();
    }
    
    public static String dumpObject(Object o, String nam, String ind) throws Exception {
	if(o == null) {
	    return "<null>";
	}
	StringBuilder sb = new StringBuilder();
	Class<?> clz = o.getClass();	// assume never null
	String clznam = clz.getName();
	if(clznam.startsWith("java")) {
	    return ind + clznam + "\n";
	}
	if(_ocache.get(clznam) != null) {
	    return ind + "-- existing -- " + clznam + "\n";
	}
	_ocache.put(clznam, "X");
	//
	// Skip all java.lang.* classes
	//
	if(clznam.startsWith("java.lang.")) {
	    return "";
	}
	if((nam == null) || (nam.length() == 0)) {
	    sb.append(ind);
	    sb.append(clz.getName());
	    sb.append("\n");
	}
	String nxtind = ind + "|   ";
	Iterable<Field> flds = getAllFields(o.getClass(), ind);
	for(Field f : flds) {
	    f.setAccessible(true);	// Make sure we can access the field
	    //
	    // Get some characteristics about this field's class
	    //
	    boolean is_array = f.getType().isArray();
	    String typ = f.getType().getTypeName();
	    String fldnam = f.getName();
	    boolean is_builtin = false;
	    is_builtin = typ.equals("byte") ||
		typ.equals("char") ||
		typ.equals("int") ||
		typ.equals("long") ||
		typ.equals("float") ||
		typ.equals("double") ||
		typ.equals("boolean");
	    Object obj = f.get(o);
	    int flags = f.getModifiers();
	    if((flags & Modifier.STATIC) == 0) {
		if(is_array) {
		    String elmtyp = f.getType().getComponentType().getName();
		    is_builtin = elmtyp.equals("byte") ||
			elmtyp.equals("char") ||
			elmtyp.equals("int") ||
			elmtyp.equals("long") ||
			elmtyp.equals("float") ||
			elmtyp.equals("double") ||
			elmtyp.equals("boolean");
		    obj = f.getType().getComponentType();
		    sb.append(nxtind);
		    sb.append(typ);
		    sb.append("\t");
		    sb.append(fldnam);
		    sb.append("\n");
		    if(!is_builtin) {
			sb.append(dumpObject(obj, fldnam, nxtind));
		    }
		} else {
		    if(is_builtin) {
			sb.append(nxtind);
			sb.append(typ);
			sb.append("\t");
			sb.append(fldnam);
			sb.append("\n");
		    } else {
			sb.append(nxtind);
			sb.append(typ);
			sb.append("\t");
			sb.append(fldnam);
			sb.append("\n");
			if(obj != null) {
			    sb.append(dumpObject(obj, fldnam, nxtind));
			}
		    }
		}
	    } else {
		//say("dumpObject() -- field " + fldnam + " is a static property");
	    }
	}
	return sb.toString();
    }
    // {{ Constructors }}
    
    
    // {{ Instance methods }}
    
}
