/* WSGui.java 
 *
 * Copyright (c) 2017 Eddy Vanlerberghe.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of Eddy Vanlerberghe shall not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EDDY VANLERBERGHE ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL EDDY VANLERBERGHE BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package evilwan;

import javax.swing.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.math.*;
import javax.xml.bind.*;
import java.beans.*;
import java.lang.reflect.*;
import javax.swing.table.*;

/**
 * Websocket extension interaction interface
 */
public class WSGui extends JPanel implements ActionListener, PropertyChangeListener {

    // {{ Class members }}
    private static String DEFAULT_FILTERCLASS = "evilwan.EvilFilter";
    //
    // "start intercept" means "when clicked then start intercepting" so currently intercept is off
    //
    private static String TOGGLE_BTN_START_INTERCEPT = "Intercept Websocket Messages Off";
    private static String TOGGLE_BTN_STOP_INTERCEPT = "Intercept Websocket Messages On";
    
    // {{ Instance members }}
    
    private JCheckBox _chk_index = null;
    private JCheckBox _chk_url = null;
    private JCheckBox _chk_dir = null;
    private JCheckBox _chk_time = null;
    private JCheckBox _chk_comment = null;
    private JCheckBox _chk_orgmsg = null;
    private JCheckBox _chk_automodmsg = null;
    private JCheckBox _chk_edtmsg = null;

    private ButtonGroup _btngrp = null;
    private JRadioButton _rad_raw = null;
    private JRadioButton _rad_b64 = null;
    private JRadioButton _rad_hex = null;
    private Object       _selected_fmt = null;
    
    private JButton _btn_save = null;

    private JTextField _txt_classname = null;
    private JButton _btn_intercept = null;
    private boolean _is_intercepting = false;
    private boolean _is_initialized = false;

    //
    //  Burp top level application frame
    //
    private Frame _burpframe = null;
    //
    // Component used for "Intercept" tab
    //
    private JPanel _inttab = null;
    //
    // Label used to show intercepted message description
    //
    private JLabel _msglbl = null;
    //
    // "Forward" button in that pane
    //
    private JButton _fwdbtn = null;
    //
    // "Intercept on/off" toggle button in that pane
    //
    private JToggleButton _togglebtn = null;
    //
    // JTabbedPane holding the "Hex" editing tab
    //
    private JTabbedPane _hextabpane = null;
    //
    // "Hex" tab for hex-editing of intercepted message
    //
    private Container _hextab = null;
    //
    // JTable in that hex-editing tab
    //
    private JTable _jtab = null;
    //
    // TableModel used by _jtab
    //
    private TableModel _model = null;
    //
    // Field in _model that holds the actual byte[] for the message to edit
    //
    private Field _data = null;
    //
    // Filter object for websocket data: this object is (re)created every time
    // that the user clicks on the start intercepting websocket traffic button.
    // This way, a user can enter the name of another filter class and start using
    // that new filter without having to leave Burp.
    //
    private WSFilter _filter = null;

    // {{ Class methods }}

    // {{ Constructors }}
    
    /**
     * Default public constructor.
     */
    public WSGui()  {
	setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
	add(new JLabel("Select Output Fields:"));
	add(new JLabel(" "));
	_chk_index = new JCheckBox("Index", true);
        add(_chk_index);
	_chk_url = new JCheckBox("URL", true);
        add(_chk_url);
	_chk_dir = new JCheckBox("Direction", true);
        add(_chk_dir);
	_chk_time = new JCheckBox("Time", false);
        add(_chk_time);
	_chk_comment = new JCheckBox("Comment", false);
        add(_chk_comment);
	_chk_orgmsg = new JCheckBox("Original Message", true);
        add(_chk_orgmsg);
	_chk_automodmsg = new JCheckBox("Auto Modified  Message", true);
        add(_chk_automodmsg);
	_chk_edtmsg = new JCheckBox("Edited Message", true);
        add(_chk_edtmsg);

	add(new JLabel(" "));
	add(new JLabel("Select Output Field Format:"));
	add(new JLabel(" "));
	_btngrp = new ButtonGroup();
	_rad_raw = new JRadioButton("Raw");
	_btngrp.add(_rad_raw);
	_rad_raw.addActionListener(this);
	add(_rad_raw);
	_rad_b64 = new JRadioButton("Base64", true);
	_btngrp.add(_rad_b64);
	_rad_b64.addActionListener(this);
	_selected_fmt = _rad_b64;
	add(_rad_b64);
	_rad_hex = new JRadioButton("Hexadecimal");
	_btngrp.add(_rad_hex);
	_rad_hex.addActionListener(this);
	add(_rad_hex);

	add(new JLabel(" "));
	_btn_save = new JButton("Save to file");
	_btn_save.addActionListener(this);
	add(_btn_save);

	//
	// Not a nice way to build a user interface but it gets the job done...
	//

	add(new JLabel(" "));
	add(Box.createVerticalGlue());

	JPanel ttt = new JPanel();
	JPanel t4 = new JPanel();
	t4.setLayout(new BoxLayout(t4, BoxLayout.Y_AXIS));

	JPanel tmp = new JPanel();
	tmp.setLayout(new FlowLayout(FlowLayout.LEFT));
	JLabel lbl = new JLabel("Filter classname:");
	tmp.add(lbl);
	_txt_classname = new JTextField(DEFAULT_FILTERCLASS, 30);
	tmp.add(_txt_classname);
	t4.add(tmp);
	_btn_intercept = new JButton(TOGGLE_BTN_START_INTERCEPT);
	_btn_intercept.addActionListener(this);
	t4.add(_btn_intercept);

	ttt.add(t4);
	add(ttt);

	add(Box.createVerticalGlue());
    }
    
    // {{ Instance methods }}
    
    public void actionPerformed(ActionEvent e) {
	if(e.getSource().equals(_btn_save)) {
	    saveToFile();
	} else if(e.getSource().equals(_btn_intercept)) {
	    toggleWSIntercept();
	} else if(e.getSource().equals(_rad_raw)) {
	    _selected_fmt = _rad_raw;
	} else if(e.getSource().equals(_rad_b64)) {
	    _selected_fmt = _rad_b64;
	} else if(e.getSource().equals(_rad_hex)) {
	    _selected_fmt = _rad_hex;
	}
    }

    private void toggleWSIntercept() {
	if(_is_intercepting) {
	    if(_togglebtn.getText().indexOf(" on") >= 0) {
		doClick(_togglebtn);
	    }
	    //
	    // Clear filter object so that user can change the classname
	    // and use the new class for filtering when she next clicks on start
	    // intercepting
	    //
	    _filter = null;
	    _btn_intercept.setText(TOGGLE_BTN_START_INTERCEPT);
	} else {
	    try {
		if(!_is_initialized) {
		    initEnvironment();
		    _is_initialized = true;
		}
		if(_filter == null) {
		    //
		    // (Re)Create filter object
		    //
		    _filter = (WSFilter) Class.forName(_txt_classname.getText()).newInstance();
		}
		Goewie.selectNamedTabPane(_hextabpane, "Hex");
		if(_togglebtn.getText().indexOf(" on") < 0) {
		    doClick(_togglebtn);
		}
	    } catch(Exception e) {
		System.out.println("caught: " + e);
		e.printStackTrace();
		System.exit(1);
	    }
	    _btn_intercept.setText(TOGGLE_BTN_STOP_INTERCEPT);
	}
	_is_intercepting = !_is_intercepting;
    }

    private void saveToFile() {
	//
	// Tricky part: analyze Burp environment in order to find the actual
	// websocket packet information (really need to get the TableModel instance
	// not just the names of classes and methods)
	//
	Object wslst = analyzeBurpEnvironment();
	//
	// Get name of output file to create
	//
	JFileChooser fildlg = new JFileChooser();
	int res = fildlg.showSaveDialog(null);
	if(res == JFileChooser.APPROVE_OPTION) {
	    dumpWSMessages(fildlg.getSelectedFile(), wslst);
	}
    }
    /**
     * Look (recursively) in the container and all its components for a JTable
     * that has a table with the specifiec name.
     */
    private static JTable lookForWaldo(Container box, String waldo) {
	JTable jt = null;
	//
	// Check if we are the one
	//
	if(box instanceof JTable) {
	    JTable jtbox = (JTable) box;
	    int n = jtbox.getColumnCount();
	    for(int i = 0; i < n; ++i) {
		String colnam = jtbox.getColumnName(i);
		if(colnam.equals(waldo)) {
		    return jtbox;
		}
	    }
	}
	//
	// Ok, not the one: descend one level and try again with each kid
	//
	Component[] parts = box.getComponents();
	for(int i = 0; i < parts.length; ++i) {
	    if(parts[i] instanceof Container) {
		jt = lookForWaldo((Container) parts[i], waldo);
		if(jt != null) {
		    return jt;
		}
	    }
	}
	return jt;
    }
    /**
     * Run-time analysis of Burp internals.
     */
    private Object analyzeBurpEnvironment() {
	//
	// Find List-type object from the GUI JTable TableModel that contains the
	// actual websocket messages info.
	//
	Object wslst = null;
	try {
	    JTable jt = null;
	    //
	    // Loop over all top-level frames and scan each recursively until we
	    // find a JTable that has a column named "Direction" (because that is the
	    // JTable that contains the websocket info: there is only one JTable with
	    // that column in the entire interface)
	    //
	    Frame[] frames = Frame.getFrames();
	    for(int i = 0; i < frames.length; ++i) {
		//
		// Only one JTable in entire GUI with a column named "Direction"...
		//
		jt = lookForWaldo(frames[i], "Direction");
		if(jt != null) {
		    //
		    // Gotcha!
		    //
		    break;
		}
	    }
	    wslst = jt.getModel();	// is a burp.e8g in v1.7.15
	} catch (Exception e) {
	    System.out.println("## caught: " + e);
	    e.printStackTrace();
	    System.exit(1);
	}
	//
	// Complete analyzis of Burp environment to find accessor methods for
	// encapsulated websocket messages
	//
	return WSPacket.init(wslst);
    }
    private void dumpBar(FileWriter os, byte[] bar) throws Exception {
	if((_selected_fmt == null) || (bar == null)) {
	    return;
	} else if(_selected_fmt.equals(_rad_raw)) {
	    os.write("<![CDATA[");
	    os.write(new String(bar));
	    os.write("]]>");
	} else if(_selected_fmt.equals(_rad_b64)) {
	    os.write("<![CDATA[");
	    os.write(DatatypeConverter.printBase64Binary(bar));
	    os.write("]]>");
	} else if(_selected_fmt.equals(_rad_hex)) {
	    os.write(new BigInteger(1, bar).toString(16));
	} else {
	    return;
	}
    }

    /**
     * Save formatted websocket messages to file.
     *
     * @param f file to create
     * @param wslst Collection of intercepted websocket message encapsulations
     */
    private void dumpWSMessages(File f, Object wslst) {
	if(f == null) {
	    //
	    // No file selected/entered
	    //
	    return;
	}
	FileWriter os = null;
	try {
	    os = new FileWriter(f);
	    os.write("<?xml version=\"1.0\"?>");
	    os.write("<wspackets>");
	    //
	    // Note: intentionally not using a more Java-esque iteration approach here
	    //
	    int i = 0;
	    for(Object o: (Collection) wslst) {
		WSPacket wp = new WSPacket(o);
		os.write("<packet>");
		if(_chk_index.isSelected()) {
		    os.write("<index>" + (i+1) + "</index>");
		}
		if(_chk_url.isSelected()) {
		    String url = wp.getURL();
		    os.write("<url>");
		    if(url != null) {
			os.write("<![CDATA[");
			os.write(url);
			os.write("]]>");
		    }
		    os.write("</url>");
		}
		if(_chk_dir.isSelected()) {
		    os.write("<direction>" + (wp.getClientToServer()? "C2S":"S2C") + "</direction>");
		}
		if(_chk_time.isSelected()) {
		    os.write("<time>" + wp.getTimestamp() + "</time>");
		}
		if(_chk_comment.isSelected()) {
		    os.write("<comment>");
		    String cmnt = wp.getComment();
		    if(cmnt != null) {
			os.write("<![CDATA[");
			os.write(cmnt);
			os.write("]]>");
		    }
		    os.write("</comment>");
		}
		if(_chk_orgmsg.isSelected()) {
		    os.write("<original_msg>");
		    dumpBar(os, wp.getOriginalMessage());
		    os.write("</original_msg>");
		}
		if(_chk_automodmsg.isSelected()) {
		    os.write("<automodified_msg>");
		    dumpBar(os, wp.getAutoModifiedMessage());
		    os.write("</automodified_msg>");
		}
		if(_chk_edtmsg.isSelected()) {
		    os.write("<edited_msg>");
		    dumpBar(os, wp.getEditedMessage());
		    os.write("</edited_msg>");
		}
		os.write("</packet>");
		++i;
	    }
	    os.write("</wspackets>");
	    os.close();
	    os = null;
	} catch(Exception e) {
	    System.out.println("## caught: " + e);
	    e.printStackTrace();
	} finally {
	    if(os != null) {
		try {
		    os.close();
		    os = null;
		} catch(Exception e) {
		    // NOP for now
		}
	    }
	}
    }
    //
    // Analyze the Burp intercept GUI components to find what we need
    //
    private void initEnvironment() throws Exception {
	_burpframe = null;
	_inttab = null;
	Frame[] frames = Frame.getFrames();
	for(int i = 0; i < frames.length; ++i) {
	    //
	    // Find component used for "Intercept" tab
	    //
	    _inttab = (JPanel)Goewie.findNamedTabPane(frames[i], "Intercept");
	    if(_inttab != null) {
		_burpframe = frames[i];
		break;
	    }
	}
	//
	// Find label used to show intercepted message description
	//
	_msglbl = Goewie.findMessageLabel(_inttab);
	//
	// Find "Forward" button in that pane
	//
	_fwdbtn = (JButton) Goewie.findNamedButton(_inttab, "Forward");
	_fwdbtn.addPropertyChangeListener(this);
	//
	// Find "Intercept On/Off" button in that pane
	//
	_togglebtn = (JToggleButton) Goewie.findNamedButton(_inttab, "Intercept");
	//
	// Find JTabbedPane holding the "Hex" editing tab
	//
	_hextabpane = Goewie.findTabbedPane(_inttab, "Hex");
	//
	// Find "Hex" tab for hex-editing of intercepted message
	//
	_hextab = (Container)Goewie.findNamedTabPane(_hextabpane, "Hex");
	//
	// Find JTable in that hex-editing tab
	//
	_jtab = Goewie.findJTable(_hextab);
	//
	// Find byte[] field containing the actual message data
	//
	_model = _jtab.getModel();
	getDataField();
    }
    //
    // Locate first (only) property in TableModel that is of type byte[]: that is the property that
    // seems to hold the raw bytes of the message
    //
    private void getDataField() throws Exception {
	Class c = _model.getClass();
	Field[] flds = c.getDeclaredFields();
	_data = null;
	for(int i = 0; i < flds.length; ++i) {
	    flds[i].setAccessible(true);	// Make sure we can access the field
	    if(flds[i].getType().equals(byte[].class)) {
		_data = flds[i];
		return;
	    }
	}
    }
    //
    // PropertyChangeListener interface
    //
    public void propertyChange(PropertyChangeEvent e) {
	if(_is_intercepting) {
	    String propnam = e.getPropertyName();
	    if(propnam.equals("enabled")) {
		boolean newval = ((Boolean)e.getNewValue()).booleanValue();
		if(newval) {
		    String msg = _msglbl.getText();
		    if((msg == null) || !msg.startsWith("WebSockets")) {
			//
			// Nothing for websockets, so fake a click on the forward button
			//
			doClick(_fwdbtn);
			return;
		    } else {
			filterWS(msg);
			doClick(_fwdbtn);
		    }
		}
	    }
	}
    }
    private void filterWS(String msg) {
	try {
	    boolean c2s = (msg.indexOf(" from ") < 0);
	    int j = msg.lastIndexOf(' ');
	    byte[] bar = getBytes();
	    byte[] out = _filter.doFilter(c2s, msg.substring(j+1), bar);
	    if(out != null) {
		setBytes(out);
	    }
	} catch(Exception e) {
	    System.out.println("caught: " + e);
	    e.printStackTrace();
	}
    }

    private byte[] getBytes() throws Exception {
	return (byte[])_data.get(_model);
    }
    private void setBytes(byte[] bar) throws Exception {
	byte[] tmp = new byte[bar.length];
	System.arraycopy(bar, 0, tmp, 0, bar.length);
	_data.set(_model, tmp);
    }
    //
    // Foefel alert: need to queue the button click at the end of the AWT event queue
    //
    private void doClick(final AbstractButton btn) {
	EventQueue.invokeLater(new Runnable() {
		public void run() {
		    btn.doClick();
		}
	    }
	    );
    }
}
