#!/usr/bin/env perl
#
# Merge extracted static strings with decompiled Java source code (e.g. from procyon)
#
# Filename conventions:
#
# abc.dump -> list of static strings generated by the "-strings" option of
#             BurpExtender
# abc.procy -> output of running procyon with "-eta" option
# abc.java -> generated  by this script, based on two other files
#
# The only parameter to this script is the name of the class ("abc")
#

use strict;

sub say {
    print join(' ', @_), "\n";
}
sub croak {
    &say(@_);
    exit(1);
}
sub usage {
    &croak('usage:', $0, 'klazzname');
}


##################################################
#
# Main code starts here
#
##################################################

#
# Reality check
#
&usage() unless scalar(@ARGV) == 1;
my($constfilnam) = $ARGV[0].'.dump';
my($procyfilnam) = $ARGV[0].'.procy';
my($javafilnam) = $ARGV[0].'.java';
#
# Parse strings output
#
my(%consts) = '';
open(F, '<'.$constfilnam) || &croak('failed to open strings list:', $!);
my($l);
while($l = <F>) {
    chomp($l);
    if($l =~ /^([^=]+)=([^;]+);$/) {
	&say('found:', $1, $2);
	my($k) = $ARGV[0].'.'.$1;
	&say('k:', $k);
	$consts{$k} = $2;
    }
}
close(F);
#
# Slurp Java code to modify
#
open(F, '<'.$procyfilnam) || &croak('failed to open decompiled java source:', $!);
my($java) = '';
while($l = <F>) {
    $java .= $l;
}
close(F);
#&say("Java:\n", $java);
#
# Replace constants
#
my($ky);
foreach $ky (keys(%consts)) {
    &say('$ky: ', $ky);
    if(length($ky) <= 0) {
	next;
    }
    my($i);
    while(($i = index($java, $ky)) >= 0) {
	substr($java, $i, length($ky)) = $consts{$ky};
    }
}
open(F, '>'.$javafilnam) || &croak('failed to create', $javafilnam, ':', $!);
print F $java;
close(F);
